# Object for accessing result of asynchronous processes.
# 
#
# <tt><b>since</b></tt> UES8-08
class Ues::Platform::Process::UESFuture

  # Maximum wait time is 30 minutes
  @@MAXIMUM_WAIT_TIMEOUT = 1800 #:nodoc:
  
  # Creates new instance of Future. As parameter is accepted:
  # 
  # <tt><b>param</b></tt> initData Another instance of UESFuture or URI of asynchronous process where to look for result.
  # 
  # <tt><b>param</b></tt> resultCode Default result code. This core is used when UESFuture.get is called without result code.
  def initialize(initData, resultCode = nil)
    if (initData.java_kind_of?Ues::Platform::Process::UESFuture)
      initData = initData.getProcessUri
    end
   @processUri = UESFuture.create_uri(initData).to_java
   @defaultResultCode = resultCode
  end

  
  # Returns result identified by the given +resultCode+. 
  #
  # <tt><b>param</b></tt> resultCode Code (regular expressions may be used) of the required result. Default is +nil+ which
  # means we are waiting for result of parent process final state. In case of explicitly given +resultCode+ result
  # is being searched also in all nested (asynchronous) subprocesses. In case of +nil+ value result is returned as
  # soon as parent process is finished (does not wait and search for results of nested asynchronous subprocesses).
  # Result code can be also specified in constructor. And this default value from constructor is used only when +nil+ is
  # passed as resultCode to this method.
  #
  # <tt><b>param</b></tt> timeout Maximum wait time in seconds (default is one minute, maximum is 30 minutes)
  #
  # <tt><b>return</b></tt> Result in form of UESURI or +nil+ if result was not provided. Throws exception in case
  # the result has not been provided in the given timeout.
  def get(resultCode = nil, timeout = 60)

    if (resultCode.java_kind_of?Numeric) || (resultCode.java_kind_of?Java::java.lang.Number)
      timeout = resultCode
      resultCode = nil
    end

    if (timeout < 0) || (timeout > @@MAXIMUM_WAIT_TIMEOUT)
      remainingTime = @@MAXIMUM_WAIT_TIMEOUT
    else
      remainingTime = timeout
    end
    wait = 0.5 # First wait will be one second (in each loop value is multiplied by 2)

    # if result_code passed to method is null, try default result code from constructor
    resultCode = @defaultResultCode if resultCode.nil?
    
    list = nil
    if !resultCode.nil?
      list = Java::java.util.ArrayList.new
      list.add(resultCode)
    end

    result = nil
    processExplorer = Java::cz.ues.platform.container.UESComponentFactory.getComponent(Java::cz.ues.platform.process.UESProcessExplorer.java_class)
    
    while true
      # Try to get result
      result = processExplorer.getProcessResult(@processUri, list)
      if !result.nil?
        # Result object was provided (event matching given code or event
        # representing process final state if no explicit code is given)
        if resultCode.nil? && (result.getState == UESProcessState::FAILED || result.getState == UESProcessState::ERROR)
          raise UESProcessExplorerRTException.new(UESProcessExplorerRTException::E05P02::M16,
            "Asynchronous process was not successfully finished:\n\t{0}", error_message(result))
        elsif !result.getErrorCode.nil?
          raise UESProcessExplorerRTException.new(UESProcessExplorerRTException::E05P02::M17,
            "Expected result cannot be provided:\n\t{0}", error_message(result))          
        else
          return result.getResult
        end
      end
      # Check process progress (so we know whether it is necessary to wait further)
      progress = processExplorer.getProcessProgress(@processUri)
      if progress.getFinished >= 0
        # We have no result, but the process is finished
        break
      end
      if remainingTime <= 0
        # We have no result, process is not finished, but we are out of time
        raise 'Result has not been provided by process in the given time.'
      end
      # Wait longer in each cycle (in order not to flood server)
      wait *= 2
      if (remainingTime > wait)
        remainingTime -= wait
      else
        # If next wait is greater than remaining wait time, wait for the remaining time only
        wait = remainingTime
        remainingTime = 0
      end
      sleep(wait.to_i)
    end
    return result
  end
  
  # Returns underlining process uri.
  def getProcessUri
    return UESFuture.create_uri(@processUri)
  end
  
  # Returns string representation of underlining process uri. 
  def to_s
    @processUri.to_s
  end
  
  # Returns json representation of underlining process uri. 
  def to_json
    "\"#{@processUri.to_s}\""
  end
  
  def method_missing(method, *args, &block) #:nodoc:
    # Delegate unknown method calling to the inner UESURI object because of backward compatibility 
    if @processUri.respond_to? method
      # log warning
      scriptUri = Ues::Platform::Script::UESScriptContext.getScriptUri
      logger.warning("Script tries to use method from the UESURI object on UESFuture object. Method name: #{method}, Script uri: #{scriptUri.to_s}.")
      # send event to inner UESURI
      return @processUri.send(method, *args, &block)
    else
      # if the method is not defined either on the UESURI object
      super
    end
  end
  
  def logger #:nodoc:
    @@log ||= Ues::Platform::Commons::UESLoggerFactory.getLogger("Ues::Platform::Process::UESFuture")
  end

  def self.create_uri(uri) # :nodoc:
    if Ues.const_get(:Core).to_s == 'Ues::Core' &&
       Ues::Core.const_get(:Commons).to_s == 'Ues::Core::Commons' &&
       Ues::Core::Commons.const_get(:UESURI).to_s == 'Ues::Core::Commons::Uri::UESURI'
      return Ues::Core::Commons::Uri::UESURI.new(uri)
    else
      return Ues::Platform::Commons::Uri::UESURI.new(uri)
    end
  end

  private

  def error_message(error_event) #:nodoc:
    error_code = error_event.getErrorCode || "#{UESRuntimeException::E05100.java_class.simple_name}.#{UESRuntimeException::E05100::MNA}"
    if error_event.getMessage.nil?
      msg = Java::cz.ues.platform.process.dto.UESMessage.getInstance(error_code, 'Unexpected error')
    else
      msg = Java::cz.ues.platform.process.dto.UESMessage.getInstance(error_event.getMessage.getCode || error_code, error_event.getMessage.getMessage, *error_event.getMessage.getParameters)
    end
    return msg.to_s
  end

end
